(********************************************************************)
(*                                                                  *)
(*                                FONT8.PAS                         *)
(*                                                                  *)
(*                       Object Oriented Font Routines              *)
(*                                                                  *)
(*                                                                  *)
(*                            Copyright (c) 1999                    *)
(*                             Mark J Iuzzolino                     *)
(*                            All Rights Reserved                   *)
(*                                                                  *)
(*                                                                  *)
(*                                                                  *)
(********************************************************************)

(*
 *   Rev 1.0  1998/05/7 mji
 * Initial Revision
 *)

{$A+,B-,D+,E+,F-,G+,I+,L+,N-,P-,Q-,R-,S+,T-,V+,X+,Y+}
{$M 16384,0}

unit font8;
interface

uses ObjGfx8;

const LeftText            = 0;
const CenterText          = 1;
const RightText           = 2;
const BottomText          = 0;
const TopText             = 2;

type TFontType=(BLF,BMF,DPF,TTF,UNC,UNP);

{BLitFont, BitMapFont, DannyPaintFont,
TrueTypeFont, UniCode font, UniCode Page}
const TFontTypeStr:array[TFontType] of string[3]=('BLF','BMF','DPF',
                                                  'TTF','UNC','UNP');
type THeader = record
       ID                 : array[0..3] of char;
       BPP                : byte;
       paddington         : array[0..9] of byte;
     end;

type TDPFHeader = record
       ID                 : array[0..2] of char; {"DPF"}
       Version            : byte; {0=version 1}
       Width,Height       : byte;
       NumOfChars         : byte;
       StartingChar       : byte;
       ColourType         : byte; {0=monochrome}
       Paddington         : array[0..6] of byte;
     end;

const DefaultBLFHeader:THeader=(ID:'BLF'+#26;BPP:8;Paddington:(0,0,0,0,0,0,0,0,0,0));
const DefaultBMFHeader:THeader=(ID:'BMF'+#26;BPP:8;Paddington:(0,0,0,0,0,0,0,0,0,0));
const DefaultDPFHeader:TDPFHeader=(ID:'DPF';Version:0;Width:0;Height:0;NumOfChars:0;
                                   StartingChar:0;ColourType:0;Paddington:(0,0,0,0,0,0,0));

type TFont8 = object
       FontDataIdx        : array[char] of longint;
       CharWidthTable     : array[char] of word;
       CharHeightTable    : array[char] of word;
       Buf                : PGfx8;
       FontDataSize       : longint;
       FontData           : word;
       constructor init(var Buffer:TGfx8);
       procedure PutText(X,Y:integer; const TextString:string); virtual;
       procedure CenterTextX(Y:integer; const TextString:string); virtual;
       procedure JustifyText(Horiz,Vert:word; const TextString:string); virtual;
       function TextWidth(const TextString:string):word; virtual;
       function TextHeight(const TextString:string):word; virtual;
       function Load(const FontFile:string):boolean; virtual;
       function Save(const FontFile:string):boolean; virtual;
       function LoadFrom(const FontFile:string; Offset:longint):boolean; virtual;
       function SaveTo(const FontFile:string; Offset:longint):boolean; virtual;
       destructor done; virtual;
     end;

type PBMFont8             = ^TBMFont8;
     TBMFont8 = object(TFont8)
       constructor Init(var Buffer:TGfx8);
       procedure SetColor(colour:byte);
       procedure SetColorIdx(idx,colour:byte);
       procedure PutText(X,Y:integer; const TextString:string); virtual;
       function SaveTo(const FontFile:string; Offset:longint):boolean; virtual;
      private
       ColourTable        : array[byte] of byte;
     end;

type PBlitFont8           = ^TBlitFont8;
     TBlitFont8 = object(TFont8)
       procedure PutText(X,Y:integer; const TextString:string); virtual;
       function SaveTo(const FontFile:string; Offset:longint):boolean; virtual;
     end;

type PDPFont8             = ^TDPFont8;
     TDPFont8 = object(TBMFont8)
       Width,Height       : byte;
       NumOfChars         : word;
       StartingChar       : byte;
       constructor init(var buffer:Tgfx8);
       procedure PutText(X,Y:integer; const TextString:string); virtual;
       function LoadFrom(const FontFile:string; Offset:longint):boolean; virtual;
       function SaveTo(const FontFile:string; Offset:longint):boolean; virtual;
     end;

{function FontType(buffer:pointer):TFontType;}
implementation

uses NewFrontier,Dos;

constructor TFont8.Init;
begin
 Buf:=@Buffer;
 FontDataSize:=0;
 fillchar(FontDataIdx,sizeof(FontDataIdx),0);
 fillchar(CharWidthTable,sizeof(CharWidthTable),0);
 fillchar(CharHeightTable,sizeof(CharHeightTable),0);
 FontData:=0;
end;

procedure TFont8.PutText;
begin
 {This is a null method.  Nobody should be calling this}
{halt;}
end;

procedure TFont8.CenterTextX;
var x:integer;
begin
 x:=((Buf^.GetMaxX+1)-TextWidth(TextString)) div 2;
 PutText(x,y,TextString)
end;

procedure TFont8.JustifyText;
var x,y:integer;
begin
 if not (Horiz in [0..2]) then exit;
 if not (Vert in [0..2]) then exit;
 case Horiz of
  LeftText:x:=0;
  CenterText:x:=((Buf^.GetMaxX)-TextWidth(TextString)) div 2;
  RightText:x:=(Buf^.GetMaxX)-TextWidth(TextString);
 end;
 case Vert of
  TopText:y:=0;
  CenterText:y:=((Buf^.GetMaxY)-TextHeight(TextString)) div 2;
  BottomText:y:=(Buf^.GetMaxY)-TextHeight(TextString);
 end;
 PutText(x,y,TextString)
end;

function TFont8.TextWidth;
var tmp,size:word;
begin
 size:=0;
 for tmp:=1 to length(TextString) do
  inc(size,CharWidthTable[TextString[tmp]]);
 TextWidth:=size
end;

function TFont8.TextHeight;
var tmp,size,tmpsize:word;
begin
 size:=0;
 for tmp:=1 to length(TextString) do
  begin
   tmpsize:=CharHeightTable[TextString[tmp]];
   if (tmpsize>size) then size:=tmpsize
  end;
 TextHeight:=size
end;

function TFont8.Load;
begin
 Load:=LoadFrom(FontFile,0)
end;

function TFont8.Save;
begin
 Save:=SaveTo(FontFile,0)
end;

function TFont8.LoadFrom;
var infile:file;
    iresult:word;
    lresult:longint;
    header:THeader;
label Failure;
begin
 LoadFrom:=FALSE;
 if (fsearch(FontFile,'')='') then exit;
 {FontFile exists.. proceed to open it}
 assign(infile,FontFile);
 reset(infile,1); {set recordsize to be a byte}
 seek(infile,Offset);
 {Unload existing font if there is one}
 if (FontDataSize<>0) then Freemem32(FontData);
 blockread(infile,header,sizeof(header));
 {Uncomment next piece of code for more error checking}
 {if (iresult<>sizeof(header)) then goto Failure;}
 blockread(infile,CharWidthTable,sizeof(CharWidthTable),iresult);
 {if (iresult<>sizeof(CharWidthTable)) then goto Failure;}
 blockread(infile,CharHeightTable,sizeof(CharHeightTable),iresult);
 {if (iresult<>sizeof(CharHeightTable)) then goto Failure;}
 blockread(infile,FontDataIdx,sizeof(FontDataIdx),iresult);
 {if (iresult<>sizeof(FontDataIdx)) then goto Failure;}
 blockread(infile,FontDataSize,sizeof(FontDataSize),iresult);
 {if (iresult<>sizeof(FontDataSize)) then goto Failure;}
 getmem32(FontData,FontDataSize);
 if (FontData=0) then goto Failure;
 blockread32(infile,ptr48(FontData,0),FontDataSize,lresult);
 if (lresult<>FontDataSize) then goto Failure;
 LoadFrom:=TRUE;
Failure:
 close(infile);
end;

function TFont8.SaveTo;
begin
end;

destructor TFont8.Done;
begin
 if (FontDataSize<>0) then Freemem32(FontData);
 Buf:=NIL
end;

constructor TBMFont8.Init;
var tmp:word;
begin
 inherited init(Buffer);
 for tmp:=low(ColourTable) to high(ColourTable) do
  ColourTable[tmp]:=tmp
end;

{SetColor is used for setting the colour in single colour bitmapped font}
procedure TBMFont8.SetColor;
begin
 ColourTable[1]:=colour
end;

{SetColorIdx is used for setting the colours in a multi-colour bitmapped font}
procedure TBMFont8.SetColorIdx;
begin
 ColourTable[idx]:=colour
end;

procedure TBMFont8.PutText;
var LetterNum:word;
    ch:char;
    xx,yy:word;
    offset:longint;
    pix:byte;
begin
 if (FontData=0) then exit;
 for LetterNum:=1 to length(TextString) do
  begin
   ch:=TextString[LetterNum];
   if (CharWidthTable[ch]<>0) and (ch<>' ') then
    begin
     offset:=FontDataIdx[ch];
     for yy:=0 to CharHeightTable[ch]-1 do
      for xx:=0 to CharWidthTable[ch]-1 do
       begin
        pix:=readmem32b(ptr48(FontData,offset));
        inc(offset);
        if (pix<>0) then Buf^.putpixel(x+xx,y+yy,ColourTable[pix])
       end;
    end;
   inc(x,CharWidthTable[ch])
  end
end;

function TBMFont8.SaveTo;
var outfile:file;
    result:longint;
    header:THeader;
begin
 SaveTo:=FALSE;                         {Set to false by default}
 if (FontData=0) then exit;             {Nothing to save; abort}
 assign(outfile,FontFile);              {assign the FontFile}
 if (fsearch(FontFile,'')='') then {file doesn't exist}
  begin
  {Make a new one}
  {$I-} rewrite(outfile,1); {$I+}
   if (IOResult<>0) then exit
  end
 else {the file does exist}
  begin
   {Open existing file and seek to the correct position}
   reset(outfile,1);
   {If the user said to use an offset of -1 then just seek to the end of
    the file otherwise seek to the offset they wanted}
   if (offset=-1) then seek(outfile,filesize(outfile))
    else seek(outfile,offset)
  end;
 blockwrite(outfile,DefaultBMFHeader,sizeof(DefaultBMFHeader));
 blockwrite(outfile,CharWidthTable,sizeof(CharWidthTable));
 blockwrite(outfile,CharHeightTable,sizeof(CharHeightTable));
 blockwrite(outfile,FontDataIdx,sizeof(FontDataIdx));
 blockwrite(outfile,FontDataSize,sizeof(FontDataSize));
 blockwrite32(outfile,ptr48(FontData,0),FontDataSize,result);
 {if (result<>FontDataSize) then begin close(outfile); exit end;}
 SaveTo:=TRUE;
 close(outfile)
end;

procedure TBlitFont8.PutText;
var LetterNum:word;
    ch:char;
begin
 if (FontData=0) then exit;
 for LetterNum:=1 to length(TextString) do
  begin
   ch:=TextString[LetterNum];
   if (CharWidthTable[ch]<>0) and (ch<>' ') then
     buf^.putblit(x,y,ptr48(FontData,FontDataIdx[ch]));
   inc(x,CharWidthTable[ch])
  end
end;

function TBLitFont8.SaveTo;
var outfile:file;
    result:longint;
    header:THeader;
begin
 SaveTo:=FALSE;                         {Set to false by default}
 if (FontData=0) then exit;             {Nothing to save; abort}
 assign(outfile,FontFile);              {assign the FontFile}
 if (fsearch(FontFile,'')='') then {file doesn't exist}
  begin
  {Make a new one}
  {$I-} rewrite(outfile,1); {$I+}
   if (IOResult<>0) then exit
  end
 else {the file does exist}
  begin
   {Open existing file and seek to the correct position}
   reset(outfile,1);
   {If the user said to use an offset of -1 then just seek to the end of
    the file otherwise seek to the offset they wanted}
   if (offset=-1) then seek(outfile,filesize(outfile))
    else seek(outfile,offset)
  end;
 blockwrite(outfile,DefaultBLFHeader,sizeof(DefaultBMFHeader));
 blockwrite(outfile,CharWidthTable,sizeof(CharWidthTable));
 blockwrite(outfile,CharHeightTable,sizeof(CharHeightTable));
 blockwrite(outfile,FontDataIdx,sizeof(FontDataIdx));
 blockwrite(outfile,FontDataSize,sizeof(FontDataSize));
 blockwrite32(outfile,ptr48(FontData,0),FontDataSize,result);
 {if (result<>FontDataSize) then begin close(outfile); exit end;}
 SaveTo:=TRUE;
 close(outfile)
end;

constructor TDPFont8.init;
begin
 inherited init(buffer);
 width:=0;
 height:=0;
 NumOfChars:=0;
 StartingChar:=0;
end;

procedure TDPFont8.PutText;
var xx,xcount,ycount:word;
    offset:longint;
    LetterNum:word;
    bits,pix:byte;
    ch:char;
begin
 if (FontData=0) then exit;
 for LetterNum:=1 to length(TextString) do
  begin
   ch:=TextString[LetterNum];
   if (CharWidthTable[ch]<>0) and (ch<>' ') then
    begin
     offset:=FontDataIdx[ch];
     xcount:=CharWidthTable[ch];
     xx:=0;
     for ycount:=0 to height-1 do
      begin
       xcount:=CharWidthTable[ch];
       xx:=0;
       repeat
        if (xx and 7=0) then
         begin
          bits:=readmem32b(ptr48(FontData,offset));
          inc(offset);
         end;
        if ((bits and 128)<>0) then
         buf^.putpixel(x+xx,y+ycount,ColourTable[1]);
        inc(bits,bits);
        inc(xx);
        dec(xcount)
       until (xcount=0);
      end;
    end;
   inc(x,CharWidthTable[ch])
  end
end;

function TDPFont8.LoadFrom;
var infile:file;
    header:TDPFheader;
    iresult,tmp:word;
    lresult:longint;
    size:word;
label Failure;
begin
 LoadFrom:=FALSE;
 if (fsearch(FontFile,'')='') then exit;
 {FontFile exists.. proceed to open it}
 assign(infile,FontFile);
 reset(infile,1); {set recordsize to be a byte}
 seek(infile,Offset);
 {Unload existing font if there is one}
 if (FontDataSize<>0) then Freemem32(FontData);
 blockread(infile,header,sizeof(header));
 width:=header.width;
 height:=header.height;
 NumOfChars:=header.NumOfChars;
 if (NumOfChars=0) then NumOfChars:=256;
 StartingChar:=header.StartingChar;
 fillchar(CharWidthTable,sizeof(CharWidthTable),0);
 fillchar(CharHeightTable,sizeof(CharHeightTable),0);
 fillchar(FontDataIdx,sizeof(FontDataIdx),0);

 size:=((width+7) div 8)*height;
 FontDataSize:=size*longint(NumOfChars);

 for tmp:=StartingChar to StartingChar+NumOfChars-1 do
  begin
   CharWidthTable[char(tmp)]:=width;
   CharHeightTable[char(tmp)]:=height;
   FontDataIdx[char(tmp)]:=size*(tmp-StartingChar);
  end;

 getmem32(FontData,FontDataSize);
 if (FontData=0) then goto Failure;
 blockread32(infile,ptr48(FontData,0),FontDataSize,lresult);
 if (lresult<>FontDataSize) then goto Failure;
 LoadFrom:=TRUE;
Failure:
 close(infile)
end;

function TDPFont8.SaveTo;
begin

end;

end.
rcum.uni-mb.si/local/fontfaq/CF_TOC.HTM
www.typesolutions.com/t2k/T2KWP.HTML